"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const entity_1 = require("@expo/entity");
const wrapNativePostgresCallAsync_1 = __importDefault(require("./errors/wrapNativePostgresCallAsync"));
class PostgresEntityDatabaseAdapter extends entity_1.EntityDatabaseAdapter {
    getFieldTransformerMap() {
        return new Map([
            [
                entity_1.JSONArrayField.name,
                {
                    /**
                     * JSON array fields must be stringified before insertion using Knex.
                     * http://knexjs.org/#Schema-json
                     */
                    write: (val) => JSON.stringify(val),
                },
            ],
            [
                entity_1.MaybeJSONArrayField.name,
                {
                    /**
                     * JSON array fields must be stringified before insertion using Knex.
                     * For this field it is only an array some of the time.
                     * http://knexjs.org/#Schema-json
                     */
                    write: (val) => (Array.isArray(val) ? JSON.stringify(val) : val),
                },
            ],
        ]);
    }
    async fetchManyWhereInternalAsync(queryInterface, tableName, tableField, tableValues) {
        return await (0, wrapNativePostgresCallAsync_1.default)(() => queryInterface
            .select()
            .from(tableName)
            .whereRaw('?? = ANY(?)', [tableField, tableValues]));
    }
    applyQueryModifiersToQuery(query, querySelectionModifiers) {
        const { orderBy, offset, limit } = querySelectionModifiers;
        let ret = query;
        if (orderBy !== undefined) {
            for (const orderBySpecification of orderBy) {
                ret = ret.orderBy(orderBySpecification.columnName, orderBySpecification.order);
            }
        }
        if (offset !== undefined) {
            ret = ret.offset(offset);
        }
        if (limit !== undefined) {
            ret = ret.limit(limit);
        }
        return ret;
    }
    async fetchManyByFieldEqualityConjunctionInternalAsync(queryInterface, tableName, tableFieldSingleValueEqualityOperands, tableFieldMultiValueEqualityOperands, querySelectionModifiers) {
        let query = queryInterface.select().from(tableName);
        if (tableFieldSingleValueEqualityOperands.length > 0) {
            const whereObject = {};
            const nonNullTableFieldSingleValueEqualityOperands = tableFieldSingleValueEqualityOperands.filter(({ tableValue }) => tableValue !== null);
            const nullTableFieldSingleValueEqualityOperands = tableFieldSingleValueEqualityOperands.filter(({ tableValue }) => tableValue === null);
            if (nonNullTableFieldSingleValueEqualityOperands.length > 0) {
                for (const { tableField, tableValue } of nonNullTableFieldSingleValueEqualityOperands) {
                    whereObject[tableField] = tableValue;
                }
                query = query.where(whereObject);
            }
            if (nullTableFieldSingleValueEqualityOperands.length > 0) {
                for (const { tableField } of nullTableFieldSingleValueEqualityOperands) {
                    query = query.whereNull(tableField);
                }
            }
        }
        if (tableFieldMultiValueEqualityOperands.length > 0) {
            for (const { tableField, tableValues } of tableFieldMultiValueEqualityOperands) {
                const nonNullTableValues = tableValues.filter((tableValue) => tableValue !== null);
                query = query.where((builder) => {
                    builder.whereRaw('?? = ANY(?)', [tableField, [...nonNullTableValues]]);
                    // there was at least one null, allow null in this equality clause
                    if (nonNullTableValues.length !== tableValues.length) {
                        builder.orWhereNull(tableField);
                    }
                });
            }
        }
        query = this.applyQueryModifiersToQuery(query, querySelectionModifiers);
        return await (0, wrapNativePostgresCallAsync_1.default)(() => query);
    }
    async fetchManyByRawWhereClauseInternalAsync(queryInterface, tableName, rawWhereClause, bindings, querySelectionModifiers) {
        let query = queryInterface
            .select()
            .from(tableName)
            .whereRaw(rawWhereClause, bindings);
        query = this.applyQueryModifiersToQuery(query, querySelectionModifiers);
        return await (0, wrapNativePostgresCallAsync_1.default)(() => query);
    }
    async insertInternalAsync(queryInterface, tableName, object) {
        return await (0, wrapNativePostgresCallAsync_1.default)(() => queryInterface.insert(object).into(tableName).returning('*'));
    }
    async updateInternalAsync(queryInterface, tableName, tableIdField, id, object) {
        return await (0, wrapNativePostgresCallAsync_1.default)(() => queryInterface.update(object).into(tableName).where(tableIdField, id).returning('*'));
    }
    async deleteInternalAsync(queryInterface, tableName, tableIdField, id) {
        return await (0, wrapNativePostgresCallAsync_1.default)(() => queryInterface.into(tableName).where(tableIdField, id).del());
    }
}
exports.default = PostgresEntityDatabaseAdapter;
//# sourceMappingURL=PostgresEntityDatabaseAdapter.js.map