"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const entity_1 = require("@expo/entity");
const knex_1 = require("knex");
const nullthrows_1 = __importDefault(require("nullthrows"));
const ErrorsTestEntity_1 = __importDefault(require("../testfixtures/ErrorsTestEntity"));
const createKnexIntegrationTestEntityCompanionProvider_1 = require("../testfixtures/createKnexIntegrationTestEntityCompanionProvider");
describe('postgres errors', () => {
    let knexInstance;
    beforeAll(() => {
        knexInstance = (0, knex_1.knex)({
            client: 'pg',
            connection: {
                user: (0, nullthrows_1.default)(process.env['PGUSER']),
                password: (0, nullthrows_1.default)(process.env['PGPASSWORD']),
                host: 'localhost',
                port: parseInt((0, nullthrows_1.default)(process.env['PGPORT']), 10),
                database: (0, nullthrows_1.default)(process.env['PGDATABASE']),
            },
        });
    });
    beforeEach(async () => {
        await ErrorsTestEntity_1.default.createOrTruncatePostgresTable(knexInstance);
    });
    afterAll(async () => {
        await ErrorsTestEntity_1.default.dropPostgresTable(knexInstance);
        await knexInstance.destroy();
    });
    it('throws EntityDatabaseAdapterTransientError on Knex timeout', async () => {
        const vc = new entity_1.ViewerContext((0, createKnexIntegrationTestEntityCompanionProvider_1.createKnexIntegrationTestEntityCompanionProvider)(knexInstance));
        await ErrorsTestEntity_1.default.creator(vc)
            .setField('id', 1)
            .setField('fieldNonNull', 'hello')
            .enforceCreateAsync();
        const shortTimeoutKnexInstance = (0, knex_1.knex)({
            client: 'pg',
            connection: {
                user: (0, nullthrows_1.default)(process.env['PGUSER']),
                password: (0, nullthrows_1.default)(process.env['PGPASSWORD']),
                host: 'localhost',
                port: parseInt((0, nullthrows_1.default)(process.env['PGPORT']), 10),
                database: (0, nullthrows_1.default)(process.env['PGDATABASE']),
            },
            acquireConnectionTimeout: 1,
        });
        const vc2 = new entity_1.ViewerContext((0, createKnexIntegrationTestEntityCompanionProvider_1.createKnexIntegrationTestEntityCompanionProvider)(shortTimeoutKnexInstance));
        await expect(ErrorsTestEntity_1.default.loader(vc2).enforcing().loadByIDAsync(1)).rejects.toThrow(entity_1.EntityDatabaseAdapterTransientError);
        await shortTimeoutKnexInstance.destroy();
    });
    it('throws EntityDatabaseAdapterNotNullConstraintError when not null is violated', async () => {
        const vc = new entity_1.ViewerContext((0, createKnexIntegrationTestEntityCompanionProvider_1.createKnexIntegrationTestEntityCompanionProvider)(knexInstance));
        await expect(ErrorsTestEntity_1.default.creator(vc)
            .setField('id', 1)
            .setField('fieldNonNull', null)
            .enforceCreateAsync()).rejects.toThrow(entity_1.EntityDatabaseAdapterNotNullConstraintError);
    });
    it('throws EntityDatabaseAdapterForeignKeyConstraintError when foreign key is violated', async () => {
        const vc = new entity_1.ViewerContext((0, createKnexIntegrationTestEntityCompanionProvider_1.createKnexIntegrationTestEntityCompanionProvider)(knexInstance));
        await expect(ErrorsTestEntity_1.default.creator(vc)
            .setField('id', 1)
            .setField('fieldNonNull', 'hello')
            .setField('fieldForeignKey', 2)
            .enforceCreateAsync()).rejects.toThrow(entity_1.EntityDatabaseAdapterForeignKeyConstraintError);
    });
    it('throws EntityDatabaseAdapterUniqueConstraintError when primary key unique constraint is violated', async () => {
        const vc = new entity_1.ViewerContext((0, createKnexIntegrationTestEntityCompanionProvider_1.createKnexIntegrationTestEntityCompanionProvider)(knexInstance));
        await ErrorsTestEntity_1.default.creator(vc)
            .setField('id', 1)
            .setField('fieldNonNull', 'hello')
            .enforceCreateAsync();
        await expect(ErrorsTestEntity_1.default.creator(vc)
            .setField('id', 1)
            .setField('fieldNonNull', 'hello')
            .enforceCreateAsync()).rejects.toThrow(entity_1.EntityDatabaseAdapterUniqueConstraintError);
    });
    it('throws EntityDatabaseAdapterUniqueConstraintError when unique constraint is violated', async () => {
        const vc = new entity_1.ViewerContext((0, createKnexIntegrationTestEntityCompanionProvider_1.createKnexIntegrationTestEntityCompanionProvider)(knexInstance));
        await ErrorsTestEntity_1.default.creator(vc)
            .setField('id', 2)
            .setField('fieldNonNull', 'hello')
            .setField('fieldUnique', 'hello')
            .enforceCreateAsync();
        await expect(ErrorsTestEntity_1.default.creator(vc)
            .setField('id', 1)
            .setField('fieldNonNull', 'hello')
            .setField('fieldUnique', 'hello')
            .enforceCreateAsync()).rejects.toThrow(entity_1.EntityDatabaseAdapterUniqueConstraintError);
    });
    it('throws EntityDatabaseAdapterCheckConstraintError when check constraint is violated', async () => {
        const vc = new entity_1.ViewerContext((0, createKnexIntegrationTestEntityCompanionProvider_1.createKnexIntegrationTestEntityCompanionProvider)(knexInstance));
        await expect(ErrorsTestEntity_1.default.creator(vc)
            .setField('id', 1)
            .setField('fieldNonNull', 'hello')
            .setField('checkLessThan5', 2)
            .enforceCreateAsync()).resolves.toBeTruthy();
        await expect(ErrorsTestEntity_1.default.creator(vc)
            .setField('id', 2)
            .setField('fieldNonNull', 'hello')
            .setField('checkLessThan5', 10)
            .enforceCreateAsync()).rejects.toThrow(entity_1.EntityDatabaseAdapterCheckConstraintError);
    });
    it('throws EntityDatabaseAdapterExclusionConstraintError when exclusion constraint is violated', async () => {
        const vc = new entity_1.ViewerContext((0, createKnexIntegrationTestEntityCompanionProvider_1.createKnexIntegrationTestEntityCompanionProvider)(knexInstance));
        await expect(ErrorsTestEntity_1.default.creator(vc)
            .setField('id', 1)
            .setField('fieldNonNull', 'hello')
            .setField('fieldExclusion', 'what')
            .enforceCreateAsync()).resolves.toBeTruthy();
        await expect(ErrorsTestEntity_1.default.creator(vc)
            .setField('id', 2)
            .setField('fieldNonNull', 'hello')
            .setField('fieldExclusion', 'what')
            .enforceCreateAsync()).rejects.toThrow(entity_1.EntityDatabaseAdapterExclusionConstraintError);
    });
    it('throws EntityDatabaseAdapterUnknownError otherwise', async () => {
        const vc = new entity_1.ViewerContext((0, createKnexIntegrationTestEntityCompanionProvider_1.createKnexIntegrationTestEntityCompanionProvider)(knexInstance));
        await expect(ErrorsTestEntity_1.default.creator(vc)
            .setField('id', 1)
            .setField('fieldNonNull', 'hello')
            .setField('nonExistentColumn', 'what')
            .enforceCreateAsync()).rejects.toThrow(entity_1.EntityDatabaseAdapterUnknownError);
    });
});
//# sourceMappingURL=errors-test.js.map